'COPYRIGHT EMBEDDEDCOMPUTERS.NET 2011 - ALL RIGHTS RESERVED
'CONTACT EMAIL: contact@embeddedcomputers.net
'ANY USE OF THIS CODE MUST ADHERE TO THE LICENSE FILE INCLUDED WITH THIS SDK
'INFO: This class interfaces the SPI functionality to the BlackcatUSB hardware/firmware
'CODE: This source code class can be compiled with VS 2005 / VS 2008 / VS2010
'ACKNOWLEDGEMENT: USB driver functionality provided by LibUsbDotNet (sourceforge.net/projects/libusbdotnet) 

Imports LibUsbDotNet
Imports LibUsbDotNet.Info
Imports LibUsbDotNet.Main

'Compatible with AVR firmware BCUSB.1.09.SPI.hex
Public Class SPI_API

    'Contains SPI definition of address size and opcodes
    Public Class SpiDef

        Sub New(Optional ByVal EraseIndex As Integer = -1)
            SE_INDEX = EraseIndex
            Select Case EraseIndex
                Case -1 'Setting manually 
                Case 0
                    EraseSize = &H1000 '4KB
                Case 1
                    EraseSize = &H8000 '32KB
                Case 2
                    EraseSize = &H10000 '64KB
            End Select
        End Sub

        Private SE_INDEX As Integer
        Public USEPAGEPROGRAM As Boolean = True 'The pageprogram/write command can program the entire page or uses WORD program
        Public ChipSize As Integer = 0 'Number of bytes the chip has
        Public EraseSize As Integer = 0 'Number of bytes a sector erases
        Public PageSize As Integer = 256 'bytes
        Public AddressSize As Integer = 24 'bits
        Public JEDECID As Byte = &H9F 'Read JEDEC ID
        Public READ As Byte = &H3 'Read Data Byte
        Public FAST_READ As Byte = &HB 'Read Data Byte (high speed)
        Public WREN As Byte = &H6 'Write-Enable
        Public WRDI As Byte = &H4 'Write-Disable
        Public RDSR As Byte = &H5 'Read Status Register
        Public WRSR As Byte = &H1 'Write Status Register
        Public PAGEPROGRAM As Byte = &H2 'Page Program command
        Public WORDPROGRAM As Byte = &HAD 'AAI mode used for most SST chips
        Public SE_4KB As Byte = &H20 'Erase 4 KByte of memory array
        Public SE_32KB As Byte = &H52 'Erase 32 KByte of memory array
        Public SE_64KB As Byte = &HD8 'Erase 64 KByte of memory array
        Public SE_DEF As Byte = &H52 ' Default sector erase command for specified devices
        Public CHIPERASE As Byte = &H60
        'Number of bytes the erase command will delete
        Public ReadOnly Property SectorSize() As Integer
            Get
                Return EraseSize
            End Get
        End Property
        'Opcode to erase sector of selected chip
        Public ReadOnly Property SectorErase() As Byte
            Get
                Select Case SE_INDEX
                    Case -1
                        Return SE_DEF
                    Case 0
                        Return SE_4KB
                    Case 1
                        Return SE_32KB
                    Case 2
                        Return SE_64KB
                End Select
            End Get
        End Property
        'Number of bytes reserved for the address
        Public ReadOnly Property AddressBytes() As Integer
            Get
                Return CInt(Math.Ceiling(AddressSize / 8))
            End Get
        End Property

    End Class

    Structure SpiFlashDevice
        Dim Name As String
        Dim Found As Boolean
        Dim Manu As Byte
        Dim PartNum As UInt16
        Dim Megabits As Short
        Dim Cmd As SpiDef
    End Structure

    'Set in FlashInit
    Private intMemAddressType As Integer 'SET DeviceAddressType instead
    Private intWritePageSize As Integer

    Private FlashDB As New ArrayList 'Holds SpiFlashDevice
    Public FlashLoaded As Boolean = False
    Private MyFlashDevice As SpiFlashDevice
    Private SectorSize As Integer 'stored in KB
    Private flagSwapData As Boolean = False

    Public CustomDevice As New SpiDef 'This is what our GUI uses for a specified device
    Public UseCustom As Boolean = False

    Sub New()
        CustomDevice.ChipSize = 131072 '1Mbit
        CustomDevice.EraseSize = 32768 '32KB
        Dim SPI_4KB As New SpiDef(0)
        Dim SPI_32KB As New SpiDef(1)
        Dim SPI_64KB As New SpiDef(2)
        Dim Generic As New SpiDef(1) : Generic.SE_32KB = &HD8
        Dim SST_64KB As New SpiDef(2)
        SST_64KB.USEPAGEPROGRAM = False 'Uses WORD programming
        Generic.CHIPERASE = &HC7 'Tested on ST M25P10-A
        Dim PMC_4KB As New SpiDef(0)
        PMC_4KB.SE_4KB = &HD7
        PMC_4KB.CHIPERASE = &HC7

        AddDevice("Generic 1Mbit", &HFF, &H10, 1, Generic) 'Uses RES 3xDummy scheme
        AddDevice("Atmel AT25DF641", &H1F, &H4800, 64, SPI_32KB) 'Verified
        AddDevice("Atmel AT25DF321", &H1F, &H4700, 32, SPI_64KB) 'Verified
        AddDevice("Atmel AT25DF161", &H1F, &H4602, 16, SPI_32KB)
        AddDevice("Atmel AT25DF081", &H1F, &H4502, 8, SPI_32KB)
        AddDevice("Atmel AT25DF021", &H1F, &H4300, 2, SPI_32KB)
        AddDevice("Atmel AT26DF081A", &H1F, &H4501, 8, SPI_32KB)
        AddDevice("Atmel AT26DF161", &H1F, &H4600, 16, SPI_32KB)
        AddDevice("Atmel AT26DF161A", &H1F, &H4601, 16, SPI_32KB)
        AddDevice("Atmel AT26DF321", &H1F, &H4700, 32, SPI_64KB) 'Verified
        AddDevice("Spansion S25FL128P", &H1, &H2018, 128, SPI_64KB) 'Verified
        AddDevice("Spansion S25FL064", &H1, &H216, 64, SPI_64KB) 'Verified
        AddDevice("Spansion S25FL032", &H1, &H215, 32, SPI_64KB) 'Verified
        AddDevice("Spansion S25FL016", &H1, &H214, 16, SPI_64KB)
        AddDevice("Spansion S25FL008", &H1, &H213, 8, SPI_64KB)
        AddDevice("ST M25P128", &H20, &H2018, 128, SPI_32KB)
        AddDevice("ST M25P64", &H20, &H2017, 64, SPI_32KB) 'Verified
        AddDevice("ST M25P32", &H20, &H2016, 32, SPI_64KB)
        AddDevice("ST M25P16", &H20, &H2015, 16, SPI_64KB)
        AddDevice("ST M25P80", &H20, &H2014, 8, SPI_64KB)
        AddDevice("ST M25P40", &H20, &H2013, 4, SPI_32KB)
        AddDevice("ST M25P20", &H20, &H2012, 2, SPI_32KB)
        AddDevice("ST M25P10", &H20, &H2011, 1, Generic) 'Verified
        AddDevice("Windbond W25X40", &HEF, &H3013, 4, SPI_32KB)
        AddDevice("Windbond W25X80", &HEF, &H3014, 8, SPI_4KB)
        AddDevice("Windbond W25X16", &HEF, &H3015, 16, SPI_4KB) 'Verified
        AddDevice("Windbond W25X32", &HEF, &H3016, 32, SPI_4KB)
        AddDevice("Windbond W25X64", &HEF, &H3017, 64, SPI_4KB)
        AddDevice("Windbond W25Q64BV", &HEF, &H4017, 64, SPI_32KB)
        AddDevice("Windbond W25Q32BV", &HEF, &H4016, 32, SPI_32KB)
        AddDevice("Windbond W25Q16BV", &HEF, &H4015, 16, SPI_32KB) 'Verified
        AddDevice("Windbond W25Q80BV", &HEF, &H4014, 8, SPI_32KB) 'Verified
        AddDevice("MXIC MX25L10", &HC2, &H2011, 1, SPI_64KB) 'Verified
        AddDevice("MXIC MX25L20", &HC2, &H2012, 2, SPI_4KB)
        AddDevice("MXIC MX25L40", &HC2, &H2013, 4, SPI_64KB)
        AddDevice("MXIC MX25L80", &HC2, &H2014, 8, SPI_64KB)
        AddDevice("MXIC MX25L160", &HC2, &H2015, 16, SPI_64KB) 'Verified
        AddDevice("MXIC MX25L320", &HC2, &H2016, 32, SPI_64KB)
        AddDevice("MXIC MX25L640", &HC2, &H2017, 64, SPI_64KB) 'Verified
        AddDevice("MXIC MX25L128", &HC2, &H2018, 128, SPI_64KB)
        AddDevice("EON EN25F20", &H1C, &H3112, 2, SPI_4KB)
        AddDevice("EON EN25F40", &H1C, &H3113, 4, SPI_4KB) 'Verified - recommend fosc(4)
        AddDevice("EON EN25F80", &H1C, &H3114, 8, SPI_4KB)
        AddDevice("EON EN25P16", &H1C, &H2015, 16, SPI_32KB)
        AddDevice("EON EN25P32", &H1C, &H2016, 32, SPI_32KB)
        AddDevice("EON EN25P64", &H1C, &H2017, 64, SPI_32KB)
        AddDevice("SST 25WF040", &HBF, &H2504, 4, SST_64KB)
        AddDevice("SST 25VF040B", &HBF, &H258D, 4, SST_64KB)
        AddDevice("SST 25VF080", &HBF, &H80BF, 8, SST_64KB)
        AddDevice("SST 25VF080B", &HBF, &H258E, 8, SST_64KB) 'Verified
        AddDevice("SST 26VF016", &HBF, &H2601, 16, SST_64KB)
        AddDevice("SST 25VF016B", &HBF, &H2541, 16, SST_64KB) 'Verified
        AddDevice("SST 25VF032B", &HBF, &H2542, 32, SST_64KB)
        AddDevice("SST 25VF032B", &HBF, &H254A, 32, SST_64KB) 'Verified
        AddDevice("SST 26VF032", &HBF, &H2602, 32, SST_64KB)
        AddDevice("SST 25VF064B", &HBF, &H2543, 64, SST_64KB)
        AddDevice("SST 26VF064", &HBF, &H2603, 64, SST_64KB)
        AddDevice("SST 25VF128B", &HBF, &H2544, 128, SST_64KB)
        AddDevice("PMC PM25LV010", &H9D, &H7C7F, 1, PMC_4KB)
        AddDevice("PMC PM25LV020", &H9D, &H7D7F, 2, PMC_4KB)
        AddDevice("PMC PM25LV040", &H9D, &H7E7F, 4, PMC_4KB)
        AddDevice("PMC PM25LV080B", &H7F, &H9D13, 8, PMC_4KB)
        AddDevice("PMC PM25LV016B", &H7F, &H9D14, 16, PMC_4KB)

    End Sub

    Public Sub AddDevice(ByRef name As String, ByRef manuid As Byte, ByRef part As UInt16, ByRef megabits As Short, ByVal OPCODES As SpiDef)
        Dim n As New SpiFlashDevice
        n.Name = name
        n.Manu = manuid
        n.PartNum = part
        n.Megabits = megabits
        n.Cmd = OPCODES
        FlashDB.Add(n)
    End Sub
    'Returns true if device ID was read
    Public Function ReadDeviceID(ByRef ReturnID As Integer) As Boolean
        UpdateProgress(0)
        Dim order As SPI_ORDER
        Dim mode As SPI_MODE
        Dim clock As SPI_CLOCK
        Dim SupportedModes() As SPI_CLOCK = CType(GetSpiModes(), SPI_CLOCK())
        Dim SupportedClocks() As SPI_MODE = CType(GetSpiClocks(), SPI_MODE())
        Dim SupportedOrders() As SPI_ORDER = GetSpiOrders()
        Dim TotalModes As Integer = SupportedModes.Length * SupportedClocks.Length * SupportedOrders.Length
        Dim i As Integer = 0
        For Each order In SupportedOrders
            For Each clock In SupportedClocks
                For Each mode In SupportedModes
                    i = i + 1
                    SetDeviceConfig(clock, mode, order)
                    Application.DoEvents()
                    Sleep(100)
                    If AbortOperation Then Return False
                    ReturnID = DeviceReadId()
                    If Not ReturnID = &HFFFFFF Then
                        WriteConsole("Opening device in SPI Mode " & ModeToStr(mode) & " (Fosc/" & ClockToStr(clock) & " " & OrderToStr(order) & ")")
                        Return True
                    End If
                    UpdateProgress(CInt((i / TotalModes) * 100))
                Next
            Next
        Next
        UpdateProgress(0)
        Return False
    End Function

    Enum Usb2SpiRequests As Byte
        BC_REQ_SPI_CMD = &H10
        BC_REQ_SPI_CMD_WR = &H11
        BC_REQ_CONFIG_SPI = &H12
        BC_REQ_ABORT_SPI_CMD = &H14
        BC_REQ_SPI_CMD_PROG = &H16
        BC_REQ_SPI_CMD_WR_BLOCK = &H17
        BC_REQ_SPI_CMD_SST = &H18
        BC_REQ_LED = &H20
        BC_REQ_FIRMWARE = &H30
        BC_REQ_ECHO = &H40
    End Enum

    Enum Usb2SpiRequestParams As Short
        BC_REQ_SPI_CMD_SEND_WR_EN = &H1
        BC_REQ_SPI_CMD_WAIT_READY = &H2
    End Enum

    Private Function GetSpiOrders() As SPI_ORDER()
        If Hardcode_Order = SPI_ORDER.SPI_UNSPECIFIED Then
            Dim ret(1) As SPI_ORDER
            ret(0) = SPI_ORDER.SPI_ORDER_MSB_FIRST
            ret(1) = SPI_ORDER.SPI_ORDER_LSB_FIRST
            Return ret
        Else
            Dim ret(0) As SPI_ORDER
            ret(0) = Hardcode_Order
            Return ret
        End If
    End Function

    Private Function GetSpiClocks() As SPI_CLOCK()
        If Hardcode_Clock = SPI_CLOCK.SPI_UNSPECIFIED Then
            Dim ret(3) As SPI_CLOCK
            ret(0) = SPI_CLOCK.SPI_CLOCK_FOSC_2
            ret(1) = SPI_CLOCK.SPI_CLOCK_FOSC_4
            ret(2) = SPI_CLOCK.SPI_CLOCK_FOSC_8
            ret(3) = SPI_CLOCK.SPI_CLOCK_FOSC_16
            'ret(4) = SPI_CLOCK.SPI_CLOCK_FOSC_32
            'ret(5) = SPI_CLOCK.SPI_CLOCK_FOSC_64
            'ret(6) = SPI_CLOCK.SPI_CLOCK_FOSC_128
            Return ret
        Else
            Dim ret(0) As SPI_CLOCK
            ret(0) = Hardcode_Clock
            Return ret
        End If
    End Function

    Private Function GetSpiModes() As SPI_MODE()
        If Hardcode_Mode = SPI_MODE.SPI_UNSPECIFIED Then
            Dim ret(3) As SPI_MODE
            ret(0) = SPI_MODE.SPI_MODE_0
            ret(1) = SPI_MODE.SPI_MODE_1
            ret(2) = SPI_MODE.SPI_MODE_2
            ret(3) = SPI_MODE.SPI_MODE_3
            Return ret
        Else
            Dim ret(0) As SPI_MODE
            ret(0) = Hardcode_Mode
            Return ret
        End If
    End Function

    Private Function OrderToStr(ByVal Order As SPI_ORDER) As String
        Select Case Order
            Case SPI_ORDER.SPI_ORDER_MSB_FIRST
                Return "MSB"
            Case SPI_ORDER.SPI_ORDER_LSB_FIRST
                Return "LSB"
            Case Else
                Return "MSB"
        End Select
    End Function

    Private Hardcode_Clock As SPI_CLOCK = SPI_CLOCK.SPI_UNSPECIFIED
    Private Hardcode_Mode As SPI_MODE = SPI_MODE.SPI_UNSPECIFIED
    Private Hardcode_Order As SPI_ORDER = SPI_ORDER.SPI_UNSPECIFIED

    Public Sub SpecifyClock(ByVal Divider As Integer)
        Select Case Divider
            Case 2
                Hardcode_Clock = SPI_CLOCK.SPI_CLOCK_FOSC_2
                WriteConsole("SPI clock mode set to Fosc/2 (8 MHz)")
            Case 4
                Hardcode_Clock = SPI_CLOCK.SPI_CLOCK_FOSC_4
                WriteConsole("SPI clock mode set to Fosc/4 (4 MHz)")
            Case 8
                Hardcode_Clock = SPI_CLOCK.SPI_CLOCK_FOSC_8
                WriteConsole("SPI clock mode set to Fosc/8 (2 MHz)")
            Case 16
                Hardcode_Clock = SPI_CLOCK.SPI_CLOCK_FOSC_16
                WriteConsole("SPI clock mode set to Fosc/16 (1 MHz)")
            Case 32
                Hardcode_Clock = SPI_CLOCK.SPI_CLOCK_FOSC_32
                WriteConsole("SPI clock mode set to Fosc/32 (500 KHz)")
            Case 64
                Hardcode_Clock = SPI_CLOCK.SPI_CLOCK_FOSC_64
                WriteConsole("SPI clock mode set to Fosc/64 (250 KHz)")
            Case 128
                Hardcode_Clock = SPI_CLOCK.SPI_CLOCK_FOSC_128
                WriteConsole("SPI clock mode set to Fosc/128 (125 KHz)")
            Case 0 'Removes hardcode
                Hardcode_Clock = SPI_CLOCK.SPI_UNSPECIFIED
                WriteConsole("SPI clock set to auto configure")
            Case -1 'Removes hardcode
                Hardcode_Clock = SPI_CLOCK.SPI_UNSPECIFIED
                WriteConsole("SPI clock set to auto configure")
            Case Else
                WriteConsole("Error setting SPI clock, speed divider unknown: " & CStr(Divider))
        End Select
    End Sub

    Public Sub SpecifyOrder(ByVal Order As String)
        Select Case Order
            Case """MSB"""
                Hardcode_Order = SPI_ORDER.SPI_ORDER_MSB_FIRST
                WriteConsole("SPI bit order set to MSB")
            Case """LSB"""
                Hardcode_Order = SPI_ORDER.SPI_ORDER_LSB_FIRST
                WriteConsole("SPI bit order set to LSB")
            Case "" 'Remove
                Hardcode_Order = SPI_ORDER.SPI_UNSPECIFIED
                WriteConsole("SPI bit order set to auto configure")
            Case Else 'Error
                WriteConsole("Error setting SPI bit order: must be MSB or LSB")
        End Select
    End Sub

    Public Sub SpecifyMode(ByVal Mode As Integer)
        Select Case Mode
            Case 0
                Hardcode_Mode = SPI_MODE.SPI_MODE_0
                WriteConsole("SPI mode set to 0")
            Case 1
                Hardcode_Mode = SPI_MODE.SPI_MODE_1
                WriteConsole("SPI mode set to 1")
            Case 2
                Hardcode_Mode = SPI_MODE.SPI_MODE_2
                WriteConsole("SPI mode set to 2")
            Case 3
                Hardcode_Mode = SPI_MODE.SPI_MODE_3
                WriteConsole("SPI mode set to 3")
            Case -1 'Remove
                Hardcode_Mode = SPI_MODE.SPI_UNSPECIFIED
                WriteConsole("SPI mode set to auto configure")
            Case Else
                WriteConsole("Error setting SPI mode, mode not supported: " & CStr(Mode))
        End Select
    End Sub

    Public Property SwapData() As Boolean
        Get
            Return flagSwapData
        End Get
        Set(ByVal value As Boolean)
            flagSwapData = value
        End Set
    End Property

    Private Function ModeToStr(ByVal Mode As SPI_MODE) As String
        Select Case Mode
            Case SPI_MODE.SPI_MODE_0
                Return "0"
            Case SPI_MODE.SPI_MODE_1
                Return "1"
            Case SPI_MODE.SPI_MODE_2
                Return "2"
            Case SPI_MODE.SPI_MODE_3
                Return "3"
            Case Else
                Return "0"
        End Select
    End Function

    Private Function ClockToStr(ByVal Clk As SPI_CLOCK) As String
        Select Case Clk
            Case SPI_CLOCK.SPI_CLOCK_FOSC_2
                Return "2"
            Case SPI_CLOCK.SPI_CLOCK_FOSC_4
                Return "4"
            Case SPI_CLOCK.SPI_CLOCK_FOSC_8
                Return "8"
            Case SPI_CLOCK.SPI_CLOCK_FOSC_16
                Return "16"
            Case SPI_CLOCK.SPI_CLOCK_FOSC_32
                Return "32"
            Case SPI_CLOCK.SPI_CLOCK_FOSC_64
                Return "64"
            Case SPI_CLOCK.SPI_CLOCK_FOSC_128
                Return "128"
            Case Else
                Return "2"
        End Select
    End Function
    'Returns -1 if not detected, 0 if detected and supported, and 1 if detected and not supported
    Public Function FlashInit() As Integer
        Dim ID As Integer
        If Not ReadDeviceID(ID) Then
            If AbortOperation Then Return -1
            WriteConsole("Unable to connect to compatible SPI device")
            Return -1
        End If
        Dim MFG As Integer = (ID And &HFF0000) >> 16
        Dim PART As Integer = (ID And &HFFFF)
        Dim Found As Boolean = False
        If UseCustom Then
            MyFlashDevice = New SpiFlashDevice()
            MyFlashDevice.Name = "SPI_SPECIFIC"
            MyFlashDevice.PartNum = 0
            MyFlashDevice.Megabits = CShort(CustomDevice.ChipSize / 131072)
            MyFlashDevice.Cmd = CustomDevice
        Else 'Search for it
            Dim item As SpiFlashDevice
            For Each item In FlashDB
                If item.Manu = MFG And item.PartNum = PART Then
                    Found = True
                    MyFlashDevice = item
                    Exit For
                End If
            Next
        End If
        SpiJedecID = Hex(MFG).PadLeft(2, CChar("0")) & " " & Hex(PART).PadLeft(4, CChar("0"))
        WriteConsole("Connected to Flash (JEDEC: " & SpiJedecID & ")")
        If UseCustom Then
            WriteConsole("Using custom settings from SPI Settings page")
        ElseIf Not Found Then
            WriteConsole("For read write support email contact@embeddedcomputers.net with JEDEC ID")
            Return 1 'Found, but not supported
        End If
        LedOn()
        FlashLoaded = True
        WriteConsole("Flash detected: " & GetFlashName() & " (" & CStr(GetFlashSize()) & " bytes)")
        WriteConsole("Programming mode: SPI (high-speed)")
        Return 0 'Supported
    End Function
    'Returns the size of the flash in megabits
    Public Function GetFlashSize() As Integer
        If Not FlashLoaded Then Return -1
        Return MyFlashDevice.Megabits * 131072
    End Function

    Public Function GetFlashName() As String
        If Not FlashLoaded Then Return ""
        Return MyFlashDevice.Name
    End Function
    'Returns the total number of sectors
    Public Function GetFlashSectors() As Integer
        If Not FlashLoaded Then Return -1
        Dim secSize As Integer = MyFlashDevice.Cmd.EraseSize
        Dim totalsize As Integer = GetFlashSize()
        Return CInt(totalsize / secSize)
    End Function

    Public Function FindSectorBase(ByVal sectorInt As Integer) As UInteger
        Return CUInt(GetSectorSize() * sectorInt)
    End Function

    Public Function GetSectorSize() As Integer
        Return MyFlashDevice.Cmd.EraseSize
    End Function
    'Sends the unprotect memory command
    Public Sub UnprotectMemory()
        WriteStatusReg(0)
    End Sub

    Private Function DeviceReady() As Boolean
        Dim Stat As Byte = ReadStatusReg()
        If (Stat And 1) = 0 Then Return True
        Return False
    End Function

    Public Sub Wait()
        Do Until DeviceReady()
            Application.DoEvents()
            Sleep(50)
        Loop
    End Sub

    '#################################################
    '######        LIBUSBDOTNET CALLS
    '#################################################

#Region "Enums"

    Enum SPI_CLOCK As Byte
        SPI_CLOCK_FOSC_2 = &H80
        SPI_CLOCK_FOSC_4 = &H0
        SPI_CLOCK_FOSC_8 = &H81
        SPI_CLOCK_FOSC_16 = &H1
        SPI_CLOCK_FOSC_32 = &H82
        SPI_CLOCK_FOSC_64 = &H2 'or 0x83
        SPI_CLOCK_FOSC_128 = &H3
        SPI_UNSPECIFIED = &HFF
    End Enum

    Enum SPI_MODE As Byte
        SPI_MODE_0 = &H0
        SPI_MODE_1 = &H4
        SPI_MODE_2 = &H8
        SPI_MODE_3 = &HC
        SPI_UNSPECIFIED = &HFF
    End Enum

    Enum SPI_ORDER As Byte
        SPI_ORDER_MSB_FIRST = &H0
        SPI_ORDER_LSB_FIRST = &H20
        SPI_UNSPECIFIED = &HFF
    End Enum

    Enum LED_STATE As Byte
        LED_STATE_OFF = &H0
        LED_STATE_ON = &H2
    End Enum

    Enum LED_BLINK As Byte
        LED_BLINK_OFF = &H0
        LED_BLINK_ON = &H1
    End Enum

#End Region

    Private bcusb As UsbDevice
    Private usbFinder As UsbDeviceFinder = New UsbDeviceFinder(&H16C0, &H5DE)
    Private USB_RW_TIMEOUT As Integer = 500
    Private SPI_FLASH_PAGE_PROGRAM_TIME As Integer = 10
    Private SPI_clock_divider As Integer 'Set by updateSPIClockDivider
    Private Const NO_ID_READ As Integer = &HFFFFFF

    Public Function IsConnected() As Boolean
        If UsbDevice.AllDevices.Find(usbFinder) Is Nothing Then Return False
        Return True
    End Function

    Public Function Connect() As Boolean
        If IsConnected() Then
            If OpenDevice() Then
                If EchoTest() Then
                    Return True
                Else
                    CloseDevice()
                End If
            End If
        End If
        Return False
    End Function

    Public Function OpenDevice() As Boolean
        CloseDevice()
        If Not IsConnected() Then Return False
        bcusb = UsbDevice.OpenUsbDevice(usbFinder)
        If bcusb IsNot Nothing Then
            Dim wholeUsbDevice As IUsbDevice = TryCast(bcusb, IUsbDevice)
            If wholeUsbDevice IsNot Nothing Then
                wholeUsbDevice.SetConfiguration(1)
                wholeUsbDevice.ClaimInterface(0)
            End If
            SetDeviceConfig(SPI_CLOCK.SPI_CLOCK_FOSC_2, SPI_MODE.SPI_MODE_0)
            Return True
        End If
        Return False
    End Function

    Public Function CloseDevice() As Boolean
        If IsOpened() Then
            Dim wholeUsbDevice As IUsbDevice = TryCast(bcusb, IUsbDevice)
            If Not ReferenceEquals(wholeUsbDevice, Nothing) Then
                wholeUsbDevice.ReleaseInterface(0)
            End If
            bcusb.Close()
            bcusb = Nothing
        End If
    End Function

    Public Function IsOpened() As Boolean
        If bcusb Is Nothing Then Return False
        If bcusb.IsOpen Then Return True
        Return False
    End Function

    Public Sub EraseSector(ByVal secNum As Integer)
        Dim Addr As UInteger = FindSectorBase(secNum)
        SectorErase(Addr)
        Wait()
    End Sub
    'Sends various READ ID commands to get a response from the chip
    Public Function DeviceReadId() As Integer
        If Not IsOpened() Then Return NO_ID_READ
        Dim res As Integer = NO_ID_READ
        Dim buff As Byte() = New Byte(2) {}
        If SendSpiCmd(1, New Byte() {&H9F}, 3, buff) = 4 Then 'READ JEDEC ID (most common)
            res = (buff(0) * 65536) + (buff(1) * 256) + buff(2)
        End If
        If Not res = &HFFFFFF Then Return res
        If SendSpiCmd(4, New Byte() {&HAB, 0, 0, 0}, 3, buff) = 7 Then 'READ ID (PMC)
            res = (buff(0) * 65536) + (buff(1) * 256) + buff(2)
        End If
        If Not res = &HFFFFFF Then Return res
        ReDim buff(3)
        If SendSpiCmd(1, New Byte() {&HAB}, 4, buff, False) = 5 Then 'Read RES command
            If (buff(0) = 255) And (buff(1) = 255) And (buff(2) = 255) And (buff(3) = &H10) Then
                Return &HFF0010
            End If
        End If
        Return res
    End Function

    Public Function SetDeviceConfig(ByVal clock As SPI_CLOCK, ByVal mode As SPI_MODE, Optional ByVal order As SPI_ORDER = SPI_ORDER.SPI_ORDER_MSB_FIRST) As Boolean
        If Not IsOpened() Then Return False
        Dim spiConf As Short = CShort(CByte(clock) Or CByte(mode) Or CByte(order))
        Dim usbSetupPacket As New UsbSetupPacket(CByte(UsbCtrlFlags.Direction_Out Or UsbCtrlFlags.Recipient_Device Or UsbCtrlFlags.RequestType_Vendor), CByte(Usb2SpiRequests.BC_REQ_CONFIG_SPI), spiConf, 0, 0)
        Dim ret As Integer
        If bcusb.ControlTransfer(usbSetupPacket, Nothing, 0, ret) Then
            updateSPIClockDivider(clock)
            Return True
        End If
        Return False
    End Function

    Public Sub Disconnect()
        FlashLoaded = False
        MyFlashDevice = Nothing
        Application.DoEvents()
        'Add other disconnect stuff here
        CloseDevice()
    End Sub

    Public Sub LedOn()
        If Not IsOpened() Then Exit Sub
        Dim ledConf As Short = CShort(2) 'SOLID
        Dim usbSetupPacket As New UsbSetupPacket(CByte(UsbCtrlFlags.Direction_Out Or UsbCtrlFlags.Recipient_Device Or UsbCtrlFlags.RequestType_Vendor), CByte(Usb2SpiRequests.BC_REQ_LED), ledConf, 0, 0)
        Dim ret As Integer
        bcusb.ControlTransfer(usbSetupPacket, Nothing, 0, ret)
    End Sub

    Public Sub LedOff()
        If Not IsOpened() Then Exit Sub
        Dim ledConf As Short = CShort(0) 'OFF
        Dim usbSetupPacket As New UsbSetupPacket(CByte(UsbCtrlFlags.Direction_Out Or UsbCtrlFlags.Recipient_Device Or UsbCtrlFlags.RequestType_Vendor), CByte(Usb2SpiRequests.BC_REQ_LED), ledConf, 0, 0)
        Dim ret As Integer
        bcusb.ControlTransfer(usbSetupPacket, Nothing, 0, ret)
    End Sub

    Public Sub LedBlink()
        If Not IsOpened() Then Exit Sub
        Dim ledConf As Short = CShort(1) 'Blink
        Dim usbSetupPacket As New UsbSetupPacket(CByte(UsbCtrlFlags.Direction_Out Or UsbCtrlFlags.Recipient_Device Or UsbCtrlFlags.RequestType_Vendor), CByte(Usb2SpiRequests.BC_REQ_LED), ledConf, 0, 0)
        Dim ret As Integer
        bcusb.ControlTransfer(usbSetupPacket, Nothing, 0, ret)
    End Sub

    Public Function GetFwVersion() As String
        If Not IsOpened() Then Return Nothing
        Dim usbSetupPacket As New UsbSetupPacket(CByte(UsbCtrlFlags.Direction_In Or UsbCtrlFlags.Recipient_Device Or UsbCtrlFlags.RequestType_Vendor), CByte(Usb2SpiRequests.BC_REQ_FIRMWARE), 0, 0, 4)
        Dim ret As Integer
        Dim buff As Byte() = New Byte(4) {}
        If bcusb.ControlTransfer(usbSetupPacket, buff, 4, ret) Then
            buff(4) = buff(3)
            buff(3) = buff(2)
            buff(2) = Asc(".")
        End If
        Dim fwstr As String = BytesToString(buff)
        Return StrToSingle(fwstr).ToString
    End Function

    Public Function ReadFlash(ByVal Base As UInteger, ByVal Length As Integer) As Byte()
        Dim data(Length - 1) As Byte
        Read(Base, Length, data, CByte(MyFlashDevice.Cmd.READ))
        Return data
    End Function

    Public Sub WriteSector(ByVal secNum As Integer, ByVal data() As Byte)
        Dim Addr As UInteger = FindSectorBase(secNum)
        WriteData(Addr, data)
    End Sub

    Public Sub WriteData(ByVal Addr24 As UInteger, ByVal data() As Byte)
        If Not IsOpened() Then Exit Sub
        If flagSwapData Then SwapEndian_8bit(data)
        Dim DataLen As Integer = data.Length
        Dim ps As Integer = CInt(MyFlashDevice.Cmd.PageSize)
        Dim pm As Integer = ps - 1
        If (Addr24 And pm) <> 0 OrElse (DataLen And pm) <> 0 Then Exit Sub
        If MyFlashDevice.Cmd.USEPAGEPROGRAM Then 'We can write all bytes
            Dim cmdLen As Integer = ps + MyFlashDevice.Cmd.AddressBytes + 1
            Dim cmdCount As Integer = CInt(DataLen / ps)
            Dim cmdBlock As Byte() = New Byte(cmdLen * cmdCount - 1) {}
            Dim cmdBlockOffset As Integer = 0, dataOffset As Integer = 0
            For i As Integer = 0 To cmdCount - 1
                cmdBlockOffset = buildWrCmdBlock(MyFlashDevice.Cmd.PAGEPROGRAM, Addr24, cmdBlock, cmdBlockOffset)
                Array.Copy(data, dataOffset, cmdBlock, cmdBlockOffset, ps)
                Addr24 = CUInt(Addr24 + ps)
                dataOffset += ps
                cmdBlockOffset += ps
            Next
            SendSpiCmdWrBlock(cmdBlock, CShort(cmdLen), CShort(cmdCount))
        Else 'Can only write one byte (about 30% slower)
            Dim FirstPacket(5) As Byte
            FirstPacket(0) = MyFlashDevice.Cmd.WORDPROGRAM
            FirstPacket(1) = CByte((Addr24 And &HFF0000) >> 16)
            FirstPacket(2) = CByte((Addr24 And &HFF00) >> 8)
            FirstPacket(3) = CByte(Addr24 And &HFF)
            FirstPacket(4) = data(0)
            FirstPacket(5) = data(1)
            SendSpiCmdWrBlock(FirstPacket, CShort(FirstPacket.Length), CShort(1))
            Dim BytesLeft As Integer = data.Length - 2
            Dim CmdCount As Integer = CInt(BytesLeft / 2)
            Dim TotalSize As Integer = CInt(CmdCount + BytesLeft)
            Dim SecondPacket(TotalSize - 1) As Byte
            Dim i As Integer
            Dim bCounter As Integer = 0
            For i = 0 To TotalSize - 1 Step 3
                SecondPacket(i) = MyFlashDevice.Cmd.WORDPROGRAM
                SecondPacket(i + 1) = data(bCounter + 2)
                SecondPacket(i + 2) = data(bCounter + 3)
                bCounter = bCounter + 2
            Next
            SendSpiCmdWrBlock(SecondPacket, CShort(3), CShort(CmdCount))
            WriteDisable() 'To turn off auto program mode
            Wait()
        End If
    End Sub

    Public Sub WriteData_PerByte(ByVal Addr As UInteger, ByVal data() As Byte)
        Dim AddrSize As Integer = MyFlashDevice.Cmd.AddressBytes
        Dim CmdLen As Short = CShort(AddrSize + 2)
        Dim CmdCount As Short = CShort(data.Length)
        Dim CmdOut((CmdCount * CmdLen) - 1) As Byte
        Dim i As Integer
        Dim ByteCounter As Integer = 0
        For i = 0 To CmdCount - 1
            Dim CmdPart(CmdLen - 1) As Byte
            Dim AddrBytes() As Byte = IntToFourBytes(Addr)
            Array.Reverse(AddrBytes)
            ReDim Preserve AddrBytes(AddrSize - 1)
            Array.Reverse(AddrBytes)
            Addr = CUInt(Addr + 1)
            CmdPart(0) = MyFlashDevice.Cmd.PAGEPROGRAM 'The command to write a byte
            Array.Copy(AddrBytes, 0, CmdPart, 1, AddrBytes.Length)
            CmdPart(CmdPart.Length - 1) = data(ByteCounter) 'Byte we are writing
            ByteCounter = ByteCounter + 1
            Array.Copy(CmdPart, 0, CmdOut, i * CmdLen, CmdPart.Length)
        Next
        SendSpiCmdWrBlock(CmdOut, CmdLen, CmdCount)
    End Sub

    Private Function SectorErase(ByVal address As UInteger) As Boolean
        If Not IsConnected() Then Return False
        WriteEnable()
        Dim cmdBuff As Byte() = New Byte(MyFlashDevice.Cmd.AddressBytes) {}
        Dim n As Integer = buildCmdBuff(MyFlashDevice.Cmd.SectorErase, address, cmdBuff)
        If SendSpiCmd(CShort(n), cmdBuff, 0, Nothing, False) = n Then
            Return True
        Else
            Return False
        End If
    End Function

    Public Function BulkErase() As Boolean
        If Not IsConnected() Then Return False
        WriteEnable()
        Dim cmdBuff As Byte() = {MyFlashDevice.Cmd.CHIPERASE}
        If SendSpiCmd(1, cmdBuff, 0, Nothing, False) = 1 Then
            ISBUSYWRITING = True
            Dim t As New Stopwatch
            t.Start()
            WriteConsole("Erasing " & CStr(GetFlashSize()) & " bytes (this may take a moment)")
            Wait()
            WriteConsole("Flash erase complete in " & Format(t.ElapsedMilliseconds / 1000, "#.##") & " seconds")
            ISBUSYWRITING = False
            Return True
        Else
            Return False
        End If
    End Function

    Public Function WriteEnable() As Boolean
        If Not IsConnected() Then Return False
        Dim cmdBuff As Byte() = {MyFlashDevice.Cmd.WREN}
        If SendSpiCmd(1, cmdBuff, 0, Nothing, False) = 1 Then
            Return True
        Else
            Return False
        End If
    End Function

    Public Function WriteDisable() As Boolean
        If Not IsConnected() Then Return False
        Dim cmdBuff As Byte() = {MyFlashDevice.Cmd.WRDI}
        If SendSpiCmd(1, cmdBuff, 0, Nothing, False) = 1 Then
            Return True
        Else
            Return False
        End If
    End Function

    Public Function SendSpiCmd(ByVal wrCnt As Short, ByVal wrBuff As Byte(), ByVal rdCnt As Integer, ByVal rdBuff As Byte(), Optional ByVal sendWrEnable As Boolean = False, Optional ByVal waitForReady As Boolean = False) As Integer
        If wrCnt = 0 AndAlso rdCnt = 0 Then Return -1
        If wrCnt <> 0 AndAlso wrBuff Is Nothing Then Return -1
        If rdCnt <> 0 AndAlso rdBuff Is Nothing Then Return -1
        Dim buff As Byte() = New Byte(5) {}

        buff(0) = CByte((wrCnt And &HFF))
        buff(1) = CByte((wrCnt >> 8) And &HFF)
        buff(2) = CByte((rdCnt And &HFF))
        buff(3) = CByte((rdCnt >> 8) And &HFF)
        buff(4) = CByte((rdCnt >> 16) And &HFF)
        buff(5) = CByte((rdCnt >> 24) And &HFF)

        Dim wrCntEff As Integer = 0
        Dim rdCntEff As Integer = 0
        Dim ret As Integer
        Dim value As Short = &H0

        If sendWrEnable Then
            value = value Or CShort(Usb2SpiRequestParams.BC_REQ_SPI_CMD_SEND_WR_EN)
        End If
        If waitForReady Then
            value = value Or CShort(Usb2SpiRequestParams.BC_REQ_SPI_CMD_WAIT_READY)
        End If
        If sendWrEnable AndAlso waitForReady AndAlso wrCnt > 0 AndAlso rdCnt = 0 Then
            Dim usbSetupPacket As New UsbSetupPacket(CByte(UsbCtrlFlags.Direction_Out Or UsbCtrlFlags.Recipient_Device Or UsbCtrlFlags.RequestType_Vendor), CByte(Usb2SpiRequests.BC_REQ_SPI_CMD_PROG), value, wrCnt, wrCnt)
            If Not bcusb.ControlTransfer(usbSetupPacket, wrBuff, wrCnt, wrCntEff) Then
                Return -1
            End If
        Else
            Dim usbSetupPacket As New UsbSetupPacket(CByte(UsbCtrlFlags.Direction_Out Or UsbCtrlFlags.Recipient_Device Or UsbCtrlFlags.RequestType_Vendor), CByte(Usb2SpiRequests.BC_REQ_SPI_CMD), value, 0, 6)
            If Not bcusb.ControlTransfer(usbSetupPacket, buff, 6, ret) Then Return -1
            If wrCnt > 0 Then
                usbSetupPacket.Request = CByte(Usb2SpiRequests.BC_REQ_SPI_CMD_WR)
                usbSetupPacket.Length = wrCnt
                If Not bcusb.ControlTransfer(usbSetupPacket, wrBuff, wrCnt, wrCntEff) Then
                    Return -1
                End If
            End If
            If rdCnt > 0 Then
                Dim reader As UsbEndpointReader = bcusb.OpenEndpointReader(ReadEndpointID.Ep01)
                Dim ec As ErrorCode = reader.Read(rdBuff, 0, rdCnt, GetUsbTimeoutForByteCnt(rdCnt), rdCntEff)
                If ec <> ErrorCode.None Then Return -1
            End If
        End If
        Return wrCntEff + rdCntEff
    End Function

    Private Function GetUsbTimeoutForByteCnt(ByVal byteCnt As Integer) As Integer
        Dim spiTimeOut As Integer = CInt(byteCnt * SPI_clock_divider / 2000)
        Dim timeOut As Integer = byteCnt >> 16
        If timeOut = 0 Then timeOut = 1
        timeOut *= USB_RW_TIMEOUT
        Return timeOut + spiTimeOut
    End Function

    Private Function buildCmdBuff(ByVal cmd As Byte, ByVal address As UInteger, ByRef buff As Byte()) As Integer
        buff(0) = cmd
        If MyFlashDevice.Cmd.AddressSize = 9 Then
            If (address And &H100) <> 0 Then
                buff(0) = CByte(buff(0) Or &H8)
            End If
        End If
        Dim n As Integer = MyFlashDevice.Cmd.AddressBytes
        For i As Integer = 0 To n - 1
            buff(n - i) = CByte((address And &HFF))
            address >>= 8
        Next
        Return n + 1
    End Function

    Public Function ReadStatusReg() As Byte
        If Not IsConnected() Then Return 0
        Dim cmdBuff As Byte() = {MyFlashDevice.Cmd.RDSR}
        Dim res As Byte() = New Byte(0) {}
        SendSpiCmd(1, cmdBuff, 1, res)
        Return res(0)
    End Function

    Private Function WriteStatusReg(ByVal val As Byte) As Boolean
        If Not IsConnected() Then Return False
        Dim cmdBuff As Byte() = {MyFlashDevice.Cmd.WRSR, val}
        If SendSpiCmd(2, cmdBuff, 0, Nothing, True) = 2 Then
            Return True
        Else
            Return False
        End If
    End Function

    Private Sub updateSPIClockDivider(ByVal clock As SPI_CLOCK)
        Select Case clock
            Case SPI_CLOCK.SPI_CLOCK_FOSC_2
                SPI_clock_divider = 2
            Case SPI_CLOCK.SPI_CLOCK_FOSC_4
                SPI_clock_divider = 4
            Case SPI_CLOCK.SPI_CLOCK_FOSC_8
                SPI_clock_divider = 8
            Case SPI_CLOCK.SPI_CLOCK_FOSC_16
                SPI_clock_divider = 16
            Case SPI_CLOCK.SPI_CLOCK_FOSC_32
                SPI_clock_divider = 32
            Case SPI_CLOCK.SPI_CLOCK_FOSC_64
                SPI_clock_divider = 64
            Case SPI_CLOCK.SPI_CLOCK_FOSC_128
                SPI_clock_divider = 128
            Case Else
                SPI_clock_divider = 2
        End Select
    End Sub

    Private Function Read(ByVal address As UInteger, ByVal datalen As Integer, ByVal data As Byte(), ByVal code As Byte) As Boolean
        If Not IsOpened() Then Return False
        Dim cmdBuff As Byte() = New Byte(MyFlashDevice.Cmd.AddressBytes) {}
        Dim n As Integer = buildCmdBuff(code, address, cmdBuff)
        Dim ret As Integer = SendSpiCmd(CShort(n), cmdBuff, datalen, data)
        If flagSwapData Then
            SwapEndian_8bit(data)
        End If
        Return CBool((datalen + n))
    End Function

    Public Function Write(ByVal address As UInteger, ByVal datalen As Integer, ByVal data As Byte()) As Boolean
        If Not IsOpened() Then Return False
        If flagSwapData Then
            SwapEndian_8bit(data)
        End If
        Dim ps As Integer = MyFlashDevice.Cmd.PageSize
        Dim pm As Integer = ps - 1
        Dim cmdBuff As Byte() = New Byte(ps + MyFlashDevice.Cmd.AddressBytes) {}
        Dim wrLength As Integer
        Dim dataOffset As UInteger = 0
        If (address And pm) <> 0 Then
            wrLength = CInt(ps - (address And pm))
        Else
            wrLength = ps
        End If
        Do
            If Not wrLength < datalen Then wrLength = datalen
            Dim n As Integer = buildCmdBuff(MyFlashDevice.Cmd.PAGEPROGRAM, address + dataOffset, cmdBuff)
            Array.Copy(data, dataOffset, cmdBuff, n, wrLength)
            Dim ret As Integer = SendSpiCmd(CShort(n + wrLength), cmdBuff, 0, Nothing, True, True)
            If ret = -1 Then Return False
            datalen -= wrLength
            dataOffset = CUInt(dataOffset + wrLength)
            wrLength = ps
        Loop While datalen > 0
        Return True
    End Function

    Private Function buildWrCmdBlock(ByVal cmd As Byte, ByVal address As UInteger, ByRef buff As Byte(), ByVal buffOffset As Integer) As Integer
        buff(buffOffset) = cmd
        If MyFlashDevice.Cmd.AddressSize = 9 Then
            If (address And &H100) <> 0 Then
                buff(buffOffset) = CByte(buff(buffOffset) Or &H8)
            End If
        End If
        Dim n As Integer = MyFlashDevice.Cmd.AddressBytes
        For i As Integer = 0 To n - 1
            buff(buffOffset + n - i) = CByte(address And &HFF)
            address >>= 8
        Next
        Return buffOffset + n + 1
    End Function

    Private Function SendSpiCmdWrBlock(ByVal Data() As Byte, ByVal cmdLen As Short, ByVal cmdCount As Short) As Integer
        If Not IsOpened() Then Return -1
        Dim DataSize As Integer = Data.Length
        If Data Is Nothing OrElse cmdLen = 0 OrElse cmdCount = 0 Then Return -1
        Dim wrCntEff As Integer = 0
        Dim usbSetupPacket As New UsbSetupPacket(CByte(UsbCtrlFlags.Direction_Out Or UsbCtrlFlags.Recipient_Device Or UsbCtrlFlags.RequestType_Vendor), CByte(Usb2SpiRequests.BC_REQ_SPI_CMD_WR_BLOCK), cmdLen, cmdCount, 0)
        If Not bcusb.ControlTransfer(usbSetupPacket, Nothing, 0, wrCntEff) Then Return -1
        Dim writer As UsbEndpointWriter = bcusb.OpenEndpointWriter(WriteEndpointID.Ep02)
        Dim ec As ErrorCode = writer.Write(Data, 0, DataSize, GetUsbTimeoutForByteCnt(DataSize) + cmdCount * SPI_FLASH_PAGE_PROGRAM_TIME, wrCntEff)
        If ec <> ErrorCode.None Then Return -1
        Return wrCntEff
    End Function

    Public Function EchoTest() As Boolean
        If Not IsOpened() Then Return False
        Dim buff As Byte() = New Byte(7) {}
        Dim usbSetupPacket As New UsbSetupPacket(CByte(UsbCtrlFlags.Direction_In Or UsbCtrlFlags.Recipient_Device Or UsbCtrlFlags.RequestType_Vendor), CByte(Usb2SpiRequests.BC_REQ_ECHO), &H1234, &H5678, 8)
        Dim ret As Integer
        If bcusb.ControlTransfer(usbSetupPacket, buff, 8, ret) Then
            If buff(0) <> CByte(UsbCtrlFlags.Direction_In Or UsbCtrlFlags.Recipient_Device Or UsbCtrlFlags.RequestType_Vendor) Then
                Return False
            End If
            If buff(1) <> CByte(Usb2SpiRequests.BC_REQ_ECHO) Then
                Return False
            End If
            If buff(2) <> &H34 Then
                Return False
            End If
            If buff(3) <> &H12 Then
                Return False
            End If
            If buff(4) <> &H78 Then
                Return False
            End If
            If buff(5) <> &H56 Then
                Return False
            End If
            If buff(6) <> &H8 Then
                Return False
            End If
            If buff(7) <> &H0 Then
                Return False
            End If

            Return True
        End If
        Return False
    End Function

End Class
